from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from numpy.typing import NDArray

from pypilecore.common.piles import PileGridProperties
from pypilecore.results.soil_properties import (
    CPTTable,
    LayerTable,
    SoilProperties,
    get_soil_layer_handles,
)

Number = Union[float, int]


@dataclass(frozen=True)
class CPTTensionResultsTable:
    """Object containing the results of a single CPT."""

    pile_tip_level_nap: NDArray[np.float64]
    """The pile-tip level in [m] w.r.t. the reference."""
    A: NDArray[np.float64]
    """The area of influence of the pile, that is, the area over which
            the stress spreads around a pile within a pile group [m2]."""
    R_t_d_plug: NDArray[np.float64]
    """The root ball weight, excluding the weight of the pile (7.6.3.3 (h) NEN
            9997-1+C2:2017) [kN]."""
    R_t_d: NDArray[np.float64]
    """The design value of the tensile resistance of a pile or pile group
            (7.6.3.3 (a) NEN 9997-1+C2:2017) [kN]."""
    R_t_k: NDArray[np.float64]
    """The characteristic value of the tensile resistance of a pile or pile
            group (7.6.3.3 NEN 9997-1+C2:2017) [kN]."""
    R_t_mob_ratio: NDArray[np.float64]
    """The mobilisation ratio of the shaft bearing capacity [-]."""
    R_t_mob: NDArray[np.float64]
    """The mobilisation of the shaft bearing capacity [kN]."""
    k_v_b: NDArray[np.float64]
    """The 1-dimensional stiffness modulus at pile bottom [kN/mm]."""
    k_v_1: NDArray[np.float64]
    """The 1-dimensional stiffness modulus at pile head [MN/mm]."""
    q_s_max_mean: NDArray[np.float64]
    """The computational value of shaft friction [MPa]."""
    s_e: NDArray[np.float64]
    """The elastic shortening of the pile [mm]."""
    s_b: NDArray[np.float64]
    """The settlement of the pile bottom [mm]."""
    s_1: NDArray[np.float64]
    """The settlement of the pile [mm]."""
    sand_clay_ratio: NDArray[np.float64]
    """Contributions to the bearing capacity from sandy layers [-]."""

    def __post_init__(self) -> None:
        dict_lengths = {}
        for key, value in self.__dict__.items():
            if not np.all(np.isnan(value)):
                dict_lengths[key] = len(value)
        if len(set(dict_lengths.values())) > 1:
            raise ValueError(
                f"Inputs for LayerTable must have same lengths, but got lengths: {dict_lengths}"
            )

    @classmethod
    def from_sequences(
        cls,
        pile_tip_level_nap: Sequence[Number],
        A: Sequence[Number],
        R_t_d_plug: Sequence[Number],
        R_t_d: Sequence[Number],
        R_t_k: Sequence[Number],
        R_t_mob_ratio: Sequence[Number],
        R_t_mob: Sequence[Number],
        k_v_b: Sequence[Number],
        k_v_1: Sequence[Number],
        q_s_max_mean: Sequence[Number],
        s_e: Sequence[Number],
        s_b: Sequence[Number],
        s_1: Sequence[Number],
        sand_clay_ratio: Sequence[Number],
    ) -> CPTTensionResultsTable:
        return cls(
            pile_tip_level_nap=np.array(pile_tip_level_nap).astype(np.float64),
            A=np.array(A).astype(np.float64),
            R_t_d_plug=np.array(R_t_d_plug).astype(np.float64),
            R_t_d=np.array(R_t_d).astype(np.float64),
            R_t_k=np.array(R_t_k).astype(np.float64),
            R_t_mob_ratio=np.array(R_t_mob_ratio).astype(np.float64),
            R_t_mob=np.array(R_t_mob).astype(np.float64),
            k_v_b=np.array(k_v_b).astype(np.float64),
            k_v_1=np.array(k_v_1).astype(np.float64),
            q_s_max_mean=np.array(q_s_max_mean).astype(np.float64),
            s_e=np.array(s_e).astype(np.float64),
            s_b=np.array(s_b).astype(np.float64),
            s_1=np.array(s_1).astype(np.float64),
            sand_clay_ratio=np.array(sand_clay_ratio).astype(np.float64),
        )

    def to_pandas(self) -> pd.DataFrame:
        """Get the pandas.DataFrame representation"""
        return pd.DataFrame(self.__dict__).dropna(axis=0, how="all")


class SingleCPTTensionBearingResults:
    """
    Object that contains the results of a PileCore single-cpt calculation.

    *Not meant to be instantiated by the user.*
    """

    def __init__(
        self,
        soil_properties: SoilProperties,
        pile_head_level_nap: float,
        results_table: CPTTensionResultsTable,
        pile_grid_properties: PileGridProperties,
    ) -> None:
        """
        Parameters
        ----------
        soil_properties
            The object with soil properties
        pile_head_level_nap
            The elevation of the pile-head, in [m] w.r.t. NAP.
        results_table
            The object with CPT results.
        """
        self._sp = soil_properties
        self._pile_head_level_nap = pile_head_level_nap
        self._results_table = results_table
        self._pile_grid_properties = pile_grid_properties

    @classmethod
    def from_api_response(
        cls,
        cpt_results_dict: dict,
        ref_height: float,
        surface_level_ref: float,
        x: float | None = None,
        y: float | None = None,
    ) -> "SingleCPTTensionBearingResults":
        results_table = cpt_results_dict["results_table"]
        return cls(
            soil_properties=SoilProperties(
                cpt_table=CPTTable.from_api_response(
                    cpt_results_dict.get("cpt_chart", {})
                ),
                layer_table=LayerTable.from_api_response(
                    cpt_results_dict["layer_table"]
                ),
                test_id=cpt_results_dict.get("test_id"),
                ref_height=ref_height,
                surface_level_ref=surface_level_ref,
                groundwater_level_ref=cpt_results_dict["groundwater_level_nap"],
                x=x,
                y=y,
            ),
            pile_head_level_nap=cpt_results_dict["pile_head_level_nap"],
            results_table=CPTTensionResultsTable.from_sequences(
                pile_tip_level_nap=results_table["pile_tip_level_nap"],
                k_v_b=results_table["k_v_b"],
                k_v_1=results_table["k_v_1"],
                A=results_table["A"],
                R_t_d_plug=results_table["R_t_d_plug"],
                R_t_d=results_table["R_t_d"],
                R_t_k=results_table["R_t_k"],
                R_t_mob_ratio=results_table["R_s_mob_ratio"],
                R_t_mob=results_table["R_s_mob"],
                q_s_max_mean=results_table["q_s_max_mean"],
                s_e=results_table["s_e"],
                s_b=results_table["s_b"],
                s_1=results_table["s_1"],
                sand_clay_ratio=results_table["sand_clay_ratio"],
            ),
            pile_grid_properties=PileGridProperties.from_api_response(
                cpt_results_dict["pile_grid"]
            ),
        )

    @property
    def soil_properties(self) -> SoilProperties:
        """
        The SoilProperties object.
        """
        return self._sp

    @property
    def pile_grid_properties(self) -> PileGridProperties:
        """
        The PileGridProperties object.
        """
        return self._pile_grid_properties

    @property
    def pile_head_level_nap(self) -> float:
        """
        The elevation of the pile-head in [m] w.r.t. NAP.
        """
        return self._pile_head_level_nap

    @property
    def table(self) -> CPTTensionResultsTable:
        """The object with single-CPT results table traces."""
        return self._results_table

    def plot_bearing_capacities(
        self,
        axes: Optional[Axes] = None,
        figsize: Tuple[float, float] = (8, 10),
        add_legend: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the bearing calculation results on an `Axes' object.

        Parameters
        ----------
        axes:
            Optional `Axes` object where the bearing capacities can be plotted on.
            If not provided, a new `plt.Figure` will be activated and the `Axes`
            object will be created and returned.
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        add_legend:
            Add a legend to the second axes object
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        axes:
            The `Axes` object where the bearing capacities were plotted on.
        """

        # Create axes objects if not provided
        if axes is not None:
            if not isinstance(axes, Axes):
                raise ValueError(
                    "'axes' argument to plot_bearing_capacities() must be a `pyplot.axes.Axes` object or None."
                )
        else:
            kwargs_subplot = {
                "figsize": figsize,
                "tight_layout": True,
            }

            kwargs_subplot.update(kwargs)

            _, axes = plt.subplots(1, 1, **kwargs_subplot)

            if not isinstance(axes, Axes):
                raise ValueError(
                    "Could not create Axes objects. This is probably due to invalid matplotlib keyword arguments. "
                )

        # add horizontal lines
        axes.axhline(
            y=self.soil_properties.groundwater_level_ref,
            color="tab:blue",
            linestyle="--",
            label="Groundwater level",
        )
        axes.axhline(
            y=self.soil_properties.surface_level_ref,
            color="tab:brown",
            linestyle="--",
            label="Surface level",
        )

        # add bearing result subplot
        axes.plot(
            np.array(self.table.R_t_d),
            self.table.pile_tip_level_nap,
            color="tab:orange",
            label=r"$R_{t;d}$",
        )
        axes.plot(
            np.array(self.table.R_t_d_plug),
            self.table.pile_tip_level_nap,
            label=r"$R_{t;d;kluit}$",
            lw=3,
            color="tab:blue",
        )
        axes.set_xlabel("Force [kN]")
        axes.set_xlim(
            np.floor(np.nanmin(np.array(self.table.R_t_d) / 10.0)) * 10,
            np.ceil(np.nanmax(np.array(self.table.R_t_d) / 10.0)) * 10,
        )

        # add legend
        if add_legend:
            axes.legend(
                loc="upper left",
                bbox_to_anchor=(1, 1),
            )

        # set grid
        axes.grid()

        return axes

    def plot_bearing_overview(
        self,
        figsize: Tuple[float, float] = (10.0, 12.0),
        width_ratios: Tuple[float, float, float] = (1, 0.1, 2),
        add_legend: bool = True,
        **kwargs: Any,
    ) -> Figure:
        """
        Plot an overview of the bearing-capacities, including the .

        Parameters
        ----------
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        width_ratios:
            Tuple of width-ratios of the subplots, as the `plt.GridSpec` argument.
        add_legend:
            Add a legend to the second axes object
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        fig:
            The matplotlib Figure
        """

        kwargs_subplot = {
            "gridspec_kw": {"width_ratios": width_ratios},
            "sharey": "row",
            "figsize": figsize,
            "tight_layout": True,
        }

        kwargs_subplot.update(kwargs)

        fig, _ = plt.subplots(
            1,
            3,
            **kwargs_subplot,
        )

        ax_qc, ax_layers, ax_bearing = fig.axes
        ax_rf = ax_qc.twiny()
        assert isinstance(ax_rf, Axes)

        # Plot bearing capacities
        self.soil_properties.cpt_table.plot_qc(ax_qc, add_legend=False)
        bounds = ax_qc.get_ylim()
        self.soil_properties.cpt_table.plot_friction_ratio(ax_rf, add_legend=False)
        self.soil_properties.plot_layers(ax_layers, add_legend=False)
        ax_layers.set_ylim(bounds)

        self.plot_bearing_capacities(axes=ax_bearing, add_legend=False)

        if add_legend:
            ax_qc_legend_handles_list = ax_qc.get_legend_handles_labels()[0]
            ax_rf_legend_handles_list = ax_rf.get_legend_handles_labels()[0]
            ax_layers_legend_handles_list = get_soil_layer_handles()

            # Omit last 2 duplicate "bearing" handles
            # (groundwater_level and surface_level):
            ax_bearing_legend_handles_list = ax_bearing.get_legend_handles_labels()[0][
                2:
            ]

            handles_list = [
                *ax_qc_legend_handles_list,
                *ax_rf_legend_handles_list,
                *ax_layers_legend_handles_list,
                *ax_bearing_legend_handles_list,
            ]

            ax_bearing.legend(
                handles=handles_list,
                loc="upper left",
                bbox_to_anchor=(1, 1),
                title=(
                    "name: " + self.soil_properties.test_id
                    if self.soil_properties.test_id is not None
                    else "name: unknown"
                ),
            )

        return fig



from __future__ import annotations

from functools import lru_cache
from typing import Any, Dict, List, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.axes import Axes
from matplotlib.patches import Patch

from pypilecore.common.piles import PileProperties
from pypilecore.exceptions import UserError
from pypilecore.results.compression.single_cpt_results import (
    SingleCPTCompressionBearingResults,
)
from pypilecore.results.load_settlement import get_load_settlement_plot

Number = Union[float, int]


class CPTCompressionGroupResultsTable:
    """
    Dataclass that contains the bearing results of a group of CPTs.
    """

    def __init__(
        self,
        pile_tip_level_nap: Sequence[float],
        R_s_k: Sequence[float],
        R_b_k: Sequence[float],
        R_c_k: Sequence[float],
        R_s_d: Sequence[float],
        R_b_d: Sequence[float],
        R_c_d: Sequence[float],
        F_nk_cal_mean: Sequence[float],
        F_nk_k: Sequence[float],
        F_nk_d: Sequence[float],
        R_c_d_net: Sequence[float],
        F_c_k: Sequence[float],
        F_c_k_tot: Sequence[float],
        s_b: Sequence[float],
        s_e: Sequence[float],
        s_e_mean: Sequence[float],
        R_b_mob_ratio: Sequence[float],
        R_s_mob_ratio: Sequence[float],
        k_v_b: Sequence[float],
        k_v_1: Sequence[float],
        R_c_min: Sequence[float],
        R_c_max: Sequence[float],
        R_c_mean: Sequence[float],
        R_c_std: Sequence[float],
        R_s_mean: Sequence[float],
        R_b_mean: Sequence[float],
        var_coef: Sequence[float],
        n_cpts: Sequence[int],
        use_group_average: Sequence[bool],
        xi_normative: Sequence[str],
        xi_value: Sequence[float],
        cpt_Rc_min: Sequence[str],
        cpt_Rc_max: Sequence[str],
        cpt_normative: Sequence[str],
    ):
        """
        Parameters
        ----------
        pile_tip_level_nap:
            The pile-tip level [m] w.r.t. NAP.
        R_s_k:
            The characteristic value of the shaft bearingcapacity  [kN].
        R_b_k:
            The characteristic value of the bottom bearingcapacity [kN].
        R_c_k:
            The characteristic value of the total compressive bearingcapacity [kN].
        R_s_d:
            The design value of the shaft bearingcapacity  [kN].
        R_b_d:
            The design value of the bottom bearingcapacity [kN].
        R_c_d:
            The design value of the total bearingcapacity [kN].
        F_nk_cal_mean:
            The mean value of the calculated single-CPT negative friction forces [kN].
        F_nk_k:
            The charactertistic value of the negative friction force [kN].
        F_nk_d:
            The design value of the negative friction force [kN].
        R_c_d_net:
            The net design value of the total bearingcapacity [kN] (netto = excluding design negative friction force.).
        F_c_k:
            The characteristic value of the load on the pile head (e.g. building load) [kN]
        F_c_k_tot:
            The characteristic value of the total compressive pile load [kN] (building-load + neg. friction force).
        s_b:
            The settlement of the pile bottom [mm].
        s_e:
            The elastic shortening of the pile [mm].
        s_e_mean:
            The mean of single-CPT results for elastic shortening of the pile [mm].
        R_b_mob_ratio:
            The mobilisation ratio of the bottom bearing capacity [-].
        R_s_mob_ratio:
            The mobilisation ratio of the shaft bearing capacity [-].
        k_v_b:
            The 1-dimensional stiffness modulus at pile bottom [kN/m].
        k_v_1:
            The 1-dimensional stiffness modulus at pile head [kN/m].
        R_c_min:
            The minimum of the single-CPT values for the calculated bearingcapacity [kN].
        R_c_max:
            The maximum of the single-CPT values for the calculated bearingcapacity [kN].
        R_c_mean:
            The mean of the single-CPT values for the calculated bearingcapacity [kN].
        R_c_std:
            The standard-deviation of the single-CPT values for the calculated bearingcapacity [kN].
        R_s_mean:
            The mean of the single-CPT values for the calculated shaft bearingcapacity [kN].
        R_b_mean:
            The mean of the single-CPT values for the calculated bottom bearingcapacity [kN].
        var_coef:
            The variation coefficient [%] of the calculated bearing capacities in the group.
        n_cpts:
            The number of CPTs [-] that have been taken into account to establish the Xi value.
        use_group_average:
            If true, the group average is used for the calculation of characteristic group
            results. If false, the values of the normative CPT are used.
        xi_normative:
            The normative Xi (either Xi_3 or Xi_4)
        xi_value:
            The Xi value [-] that was applied to calculate the characteristic value of the
            total bearing capacity.
        cpt_Rc_min:
            The CPT with the lowest value for R_c_cal.
        cpt_Rc_max:
            The CPT with the highest value for R_c_cal.
        cpt_normative:
            The normative CPT. Can be "group average" if that was found to be the normative scenario.
        """
        self.pile_tip_level_nap = (
            np.array(pile_tip_level_nap).astype(np.float64).round(decimals=2)
        )
        """The pile-tip level [m] w.r.t. NAP."""
        self.R_s_k = np.array(R_s_k).astype(np.float64)
        """The characteristic value of the shaft bearingcapacity  [kN]."""
        self.R_b_k = np.array(R_b_k).astype(np.float64)
        """The characteristic value of the bottom bearingcapacity [kN]."""
        self.R_c_k = np.array(R_c_k).astype(np.float64)
        """The characteristic value of the total compressive bearingcapacity [kN]."""
        self.R_s_d = np.array(R_s_d).astype(np.float64)
        """The design value of the shaft bearingcapacity  [kN]."""
        self.R_b_d = np.array(R_b_d).astype(np.float64)
        """The design value of the bottom bearingcapacity [kN]."""
        self.R_c_d = np.array(R_c_d).astype(np.float64)
        """The design value of the total bearingcapacity [kN]."""
        self.F_nk_cal_mean = np.array(F_nk_cal_mean).astype(np.float64)
        """The mean value of the calculated single-CPT negative friction forces [kN]."""
        self.F_nk_k = np.array(F_nk_k).astype(np.float64)
        """The charactertistic value of the negative friction force [kN]."""
        self.F_nk_d = np.array(F_nk_d).astype(np.float64)
        """The design value of the negative friction force [kN]."""
        self.R_c_d_net = np.array(R_c_d_net).astype(np.float64)
        """The net design value of the total bearingcapacity [kN] (netto = excluding design negative friction force.)."""
        self.F_c_k = np.array(F_c_k).astype(np.float64)
        """The characteristic value of the load on the pile head (e.g. building load) [kN]"""
        self.F_c_k_tot = np.array(F_c_k_tot).astype(np.float64)
        """The characteristic value of the total compressive pile load [kN] (building-load + neg. friction force)."""
        self.s_b = np.array(s_b).astype(np.float64)
        """The settlement of the pile bottom [mm]."""
        self.s_e = np.array(s_e).astype(np.float64)
        """The elastic shortening of the pile [mm]."""
        self.s_e_mean = np.array(s_e_mean).astype(np.float64)
        """The mean of single-CPT results for elastic shortening of the pile [mm]."""
        self.R_b_mob_ratio = np.array(R_b_mob_ratio).astype(np.float64)
        """The mobilisation ratio of the bottom bearing capacity [-]."""
        self.R_s_mob_ratio = np.array(R_s_mob_ratio).astype(np.float64)
        """The mobilisation ratio of the shaft bearing capacity [-]."""
        self.k_v_b = np.array(k_v_b).astype(np.float64)
        """The 1-dimensional stiffness modulus at pile bottom [kN/m]."""
        self.k_v_1 = np.array(k_v_1).astype(np.float64)
        """The 1-dimensional stiffness modulus at pile head [kN/m]."""
        self.R_c_min = np.array(R_c_min).astype(np.float64)
        """The minimum of the single-CPT values for the calculated bearingcapacity [kN]."""
        self.R_c_max = np.array(R_c_max).astype(np.float64)
        """The maximum of the single-CPT values for the calculated bearingcapacity [kN]."""
        self.R_c_mean = np.array(R_c_mean).astype(np.float64)
        """The mean of the single-CPT values for the calculated bearingcapacity [kN]."""
        self.R_c_std = np.array(R_c_std).astype(np.float64)
        """The standard-deviation of the single-CPT values for the calculated bearingcapacity [kN]."""
        self.R_s_mean = np.array(R_s_mean).astype(np.float64)
        """The mean of the single-CPT values for the calculated shaft bearingcapacity [kN]."""
        self.R_b_mean = np.array(R_b_mean).astype(np.float64)
        """The mean of the single-CPT values for the calculated bottom bearingcapacity [kN]."""
        self.var_coef = np.array(var_coef).astype(np.float64)
        """The variation coefficient [%] of the calculated bearing capacities in the group."""
        self.n_cpts = np.array(n_cpts).astype(np.int32)
        """The number of CPTs [-] that have been taken into account to establish the Xi value."""
        self.use_group_average = np.array(use_group_average).astype(np.bool_)
        """If true, the group average is used for the calculation of characteristic group
        results. If false, the values of the normative CPT are used."""
        self.xi_normative = np.array(xi_normative).astype(np.str_)
        """The normative Xi (either Xi_3 or Xi_4)"""
        self.xi_value = np.array(xi_value).astype(np.float64)
        """The Xi value [-] that was applied to calculate the characteristic value of the
        total bearing capacity."""
        self.cpt_Rc_min = np.array(cpt_Rc_min).astype(np.str_)
        """The CPT with the lowest value for R_c_cal."""
        self.cpt_Rc_max = np.array(cpt_Rc_max).astype(np.str_)
        """The CPT with the highest value for R_c_cal."""
        self.cpt_normative = np.array(cpt_normative).astype(np.str_)
        """The normative CPT. Can be "group average" if that was found to be the normative scenario."""

        for value in self.__dict__.values():
            if not len(value) == len(self.pile_tip_level_nap):
                raise ValueError(
                    "Inputs for CPTGroupResults dataclass must have same length."
                )

    @lru_cache
    def to_pandas(self) -> pd.DataFrame:
        """The pandas.DataFrame representation"""
        return pd.DataFrame(self.__dict__).dropna(axis=0, how="all")

    def plot_bearing_capacities(
        self,
        axes: Axes | None = None,
        figsize: Tuple[int, int] = (8, 10),
        add_legend: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plots the design value of the bearing capacity Rcd, based on a group of CPTs.

        Parameters
        ----------
        axes:
            Optional `Axes` object where the bearing capacity data can be plotted on.
            If not provided, a new `plt.Figure` will be activated and the `Axes`
            object will be created and returned.
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        add_legend:
            Add a legend to the `Axes` object
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        axes:
            The `Axes` object where the bearing capacities were plotted on.
        """

        # Create axes objects if not provided
        if axes is not None:
            if not isinstance(axes, Axes):
                raise TypeError(
                    "'axes' argument to plot_bearing_capacities() must be a `pyplot.axes.Axes` object or None."
                )

        else:
            kwargs_subplot = {
                "figsize": figsize,
                "tight_layout": True,
            }

            kwargs_subplot.update(kwargs)

            _, axes = plt.subplots(
                1,
                1,
                **kwargs_subplot,
            )

            if not isinstance(axes, Axes):
                raise ValueError(
                    "Could not create Axes objects. This is probably due to invalid matplotlib keyword arguments. "
                )

        axes.plot(
            self.R_c_d,
            self.pile_tip_level_nap,
            "-",
            label=r"$R_{c;d}$",
        )
        axes.plot(
            self.F_nk_d,
            self.pile_tip_level_nap,
            ":",
            label=r"$F_{s;d}$ ",
        )
        axes.set_xlabel("[kN]")
        axes.set_ylabel("[m] w.r.t. NAP")

        # set grid
        axes.grid()

        # Add legend
        if add_legend:
            axes.legend(
                loc="upper left",
                bbox_to_anchor=(1, 1),
            )
        return axes


class SingleCPTCompressionBearingResultsContainer:
    """A container that holds multiple SingleCPTBearingResults objects"""

    def __init__(self, cpt_results_dict: Dict[str, SingleCPTCompressionBearingResults]):
        """
        Parameters
        ----------
        cpt_results_dict:
            A dictionary that maps the cpt-names to SingleCPTBearingResults objects.
        """
        self._cpt_results_dict = cpt_results_dict

    @classmethod
    def from_api_response(
        cls, cpt_results_list: list, cpt_input: dict
    ) -> "SingleCPTCompressionBearingResultsContainer":
        """
        Instantiates the SingleCPTBearingResultsContainer object from the "cpts" array,
        which is returned in the response of a "compression/multiple-cpts/results" endpoint call.
        """
        return cls(
            cpt_results_dict={
                cpt_results[
                    "test_id"
                ]: SingleCPTCompressionBearingResults.from_api_response(
                    cpt_results_dict=cpt_results,
                    ref_height=cpt_input[cpt_results["test_id"]]["ref_height"],
                    surface_level_ref=cpt_input[cpt_results["test_id"]][
                        "surface_level_nap"
                    ],
                    x=cpt_input[cpt_results["test_id"]].get("location", {}).get("x"),
                    y=cpt_input[cpt_results["test_id"]].get("location", {}).get("y"),
                )
                for cpt_results in cpt_results_list
            }
        )

    def __getitem__(self, test_id: str) -> SingleCPTCompressionBearingResults:
        if not isinstance(test_id, str):
            raise TypeError(f"Expected a test-id as a string, but got: {type(test_id)}")

        return self.get_cpt_results(test_id)

    @property
    def cpt_results_dict(self) -> Dict[str, SingleCPTCompressionBearingResults]:
        """The dictionary that maps the cpt-names to SingleCPTBearingResults objects."""
        return self._cpt_results_dict

    @property
    def test_ids(self) -> List[str]:
        """The test-ids of the CPTs."""
        return list(self.cpt_results_dict.keys())

    @property
    def results(self) -> List[SingleCPTCompressionBearingResults]:
        """The computed results, as a list of SingleCPTBearingResults objects."""
        return list(self.cpt_results_dict.values())

    def get_cpt_results(self, test_id: str) -> SingleCPTCompressionBearingResults:
        """
        Returns the `SingleCPTBearingResults` object for the provided test_id.
        """

        if test_id not in self.cpt_results_dict.keys():
            raise ValueError(
                f"No Cpt-results were calculated for this test-id: {test_id}. "
                "Please check the spelling or run a new calculation for this CPT."
            )

        return self.cpt_results_dict[test_id]

    def get_results_per_cpt(self, column_name: str) -> pd.DataFrame:
        """
        Returns a pandas dataframe with a single result-item, organized per CPT
        (test-id) and pile-tip-level-nap.

        Parameters
        ----------
        column_name:
            The name of the result-item / column name of the single-cpt-results table.
        """
        if column_name not in self.to_pandas().columns or column_name in [
            "pile_tip_level_nap",
            "test_id",
        ]:
            raise ValueError("Invalid column_name provided.")

        results = pd.pivot_table(
            self.to_pandas(),
            values=column_name,
            index="pile_tip_level_nap",
            columns="test_id",
            dropna=False,
        )
        return results.sort_values("pile_tip_level_nap", ascending=False)

    @lru_cache
    def to_pandas(self) -> pd.DataFrame:
        """Returns a total overview of all single-cpt results in a pandas.DataFrame representation."""
        df_list: List[pd.DataFrame] = []

        for test_id in self.cpt_results_dict:
            df = self.cpt_results_dict[test_id].table.to_pandas()
            df = df.assign(test_id=test_id)
            df_list.append(df)

        cpt_results_df = pd.concat(df_list)
        cpt_results_df = cpt_results_df.assign(
            pile_tip_level_nap=cpt_results_df.pile_tip_level_nap.round(2)
        )

        return cpt_results_df


class MultiCPTCompressionBearingResults:
    """
    Object that contains the results of a PileCore multi-cpt calculation.

    *Not meant to be instantiated by the user.*
    """

    def __init__(
        self,
        cpt_results: SingleCPTCompressionBearingResultsContainer,
        group_results_table: CPTCompressionGroupResultsTable,
        pile_properties: PileProperties,
        gamma_f_nk: float,
        gamma_r_b: float,
        gamma_r_s: float,
        soil_load: float,
    ) -> None:
        """
        Parameters
        ----------
        cpt_results:
            The container object with single-CPT results
        group_results_table:
            The table object with CPT-group-results
        pile_properties:
            The PileProperties object
        gamma_f_nk:
            Safety factor for design-values of the negative sleeve friction force. [-]
        gamma_r_b:
            Safety factor, used to obtain design-values of the pile-tip bearingcapacity. [-]
        gamma_r_s:
            Safety factor, used to obtain design-values of the sleeve bearingcapacity. [-]
        soil_load:
            (Fictive) load on soil used to calculate soil settlement [kPa]. This is
            required and used to determine settlement of pile w.r.t. soil.
        """
        self._pp = pile_properties
        self._gamma_f_nk = gamma_f_nk
        self._gamma_r_b = gamma_r_b
        self._gamma_r_s = gamma_r_s
        self._soil_load = soil_load

        self._cpt_results = cpt_results

        self._group_results_table = group_results_table

    @classmethod
    def from_api_response(
        cls, response_dict: dict, cpt_input: dict
    ) -> "MultiCPTCompressionBearingResults":
        """
        Build the object from the response payload of the PileCore endpoint
        "/compression/multi-cpt/results".
        """
        cpt_results_dict = (
            SingleCPTCompressionBearingResultsContainer.from_api_response(
                cpt_results_list=response_dict["cpts"], cpt_input=cpt_input
            )
        )
        group_results = response_dict["group_results"]
        return cls(
            cpt_results=cpt_results_dict,
            pile_properties=PileProperties.from_api_response(
                response_dict["pile_properties"]
            ),
            group_results_table=CPTCompressionGroupResultsTable(
                pile_tip_level_nap=group_results["pile_tip_level_nap"],
                R_s_k=group_results["R_s_k"],
                R_b_k=group_results["R_b_k"],
                R_c_k=group_results["R_c_k"],
                R_s_d=group_results["R_s_d"]
                if "R_s_d" in group_results
                else np.full_like(
                    group_results["pile_tip_level_nap"], fill_value=np.nan
                ),  # For backwards compatibility with PileCore-API < 2.9.0
                R_b_d=group_results["R_b_d"]
                if "R_b_d" in group_results
                else np.full_like(
                    group_results["pile_tip_level_nap"], fill_value=np.nan
                ),  # For backwards compatibility with PileCore-API < 2.9.0,
                R_c_d=group_results["R_c_d"],
                F_nk_cal_mean=group_results["F_nk_cal_mean"],
                F_nk_k=group_results["F_nk_k"],
                F_nk_d=group_results["F_nk_d"],
                R_c_d_net=group_results["R_c_d_net"],
                F_c_k=group_results["F_c_k"],
                F_c_k_tot=group_results["F_c_k_tot"],
                s_b=group_results["s_b"],
                s_e=group_results["s_e"],
                s_e_mean=group_results["s_e_mean"],
                R_b_mob_ratio=group_results["R_b_mob_ratio"],
                R_s_mob_ratio=group_results["R_s_mob_ratio"],
                k_v_b=group_results["k_v_b"],
                k_v_1=group_results["k_v_1"],
                R_c_min=group_results["R_c_min"],
                R_c_max=group_results["R_c_max"],
                R_c_mean=group_results["R_c_mean"],
                R_c_std=group_results["R_c_std"],
                R_s_mean=group_results["R_s_mean"],
                R_b_mean=group_results["R_b_mean"],
                var_coef=group_results["var_coef"],
                n_cpts=group_results["n_cpts"],
                use_group_average=group_results["use_group_average"],
                xi_normative=group_results["xi_normative"],
                xi_value=group_results["xi_value"],
                cpt_Rc_min=group_results["cpt_Rc_min"],
                cpt_Rc_max=group_results["cpt_Rc_max"],
                cpt_normative=group_results["cpt_normative"],
            ),
            gamma_f_nk=response_dict["calculation_params"]["gamma_f_nk"],
            gamma_r_b=response_dict["calculation_params"]["gamma_r_b"],
            gamma_r_s=response_dict["calculation_params"]["gamma_r_s"],
            soil_load=response_dict["calculation_params"]["soil_load"],
        )

    @property
    def pile_properties(self) -> PileProperties:
        """
        The PileProperties object.
        """
        return self._pp

    @property
    def cpt_results(self) -> SingleCPTCompressionBearingResultsContainer:
        """The SingleCPTBearingResultsContainer object."""
        return self._cpt_results

    @property
    def cpt_names(self) -> List[str]:
        """The test-ids of the CPTs."""
        return self.cpt_results.test_ids

    @property
    def group_results_table(self) -> CPTCompressionGroupResultsTable:
        """The CPTGroupResultsTable dataclass, containing the group results."""
        return self._group_results_table

    def boxplot(
        self,
        attribute: str,
        axes: Axes | None = None,
        figsize: Tuple[float, float] = (6.0, 6.0),
        show_sqrt: bool = False,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot a box and whisker plot for a given attribute.


        .. code-block:: none

                     MIN      Q1   median  Q3       MAX
                               |-----:-----|
                      |--------|     :     |--------|
                               |-----:-----|

        Parameters
        ----------
        attribute:
            result attribute to create boxplot. Please note that the attribute name must be present in
            the `CPTResultsTable` and `CPTGroupResultsTable` class.
        axes:
            Optional `Axes` object where the boxplot data can be plotted on.
            If not provided, a new `plt.Figure` will be activated and the `Axes`
            object will be created and returned.
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        show_sqrt:
            Add sqrt(2) bandwidth to figure
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        axes:
            The `Axes` object where the settlement curves were plotted on
        """

        # validate attribute
        if (
            attribute not in self.cpt_results.results[0].table.__dict__.keys()
            or attribute not in self.group_results_table.__dict__.keys()
        ):
            raise ValueError(
                f"""
                {attribute} is not present in CPTResultsTable or CPTGroupResultsTable class.
                Please select on of the following attributes:
                {
                    set(self.cpt_results.results[0].table.__dict__.keys())
                     & set(self.group_results_table.__dict__.keys())
                }
                """
            )

        # Create axes objects if not provided
        if axes is not None:
            if not isinstance(axes, Axes):
                raise ValueError(
                    "'axes' argument to boxplot() must be a `pyplot.axes.Axes` object or None."
                )
        else:
            kwargs_subplot = {
                "figsize": figsize,
                "tight_layout": True,
            }

            kwargs_subplot.update(kwargs)

            _, axes = plt.subplots(
                1,
                1,
                **kwargs_subplot,
            )

            if not isinstance(axes, Axes):
                raise ValueError(
                    "Could not create Axes objects. This is probably due to invalid matplotlib keyword arguments. "
                )

        # Collect data from single calculation
        data = np.array(
            [
                item.table.__getattribute__(attribute)
                for item in self.cpt_results.results
            ]
        )

        # Draw a box and whisker plot
        axes.boxplot(
            np.flip(data, axis=0),
            tick_labels=np.flip(self.group_results_table.pile_tip_level_nap),
            whis=(0, 100),
            autorange=True,
            vert=False,
            patch_artist=True,
            showmeans=True,
            zorder=0,
        )

        # ad additional bandwidth of sqrt(2) of the mean value
        if show_sqrt:
            axes.scatter(
                np.flip(data.mean(axis=0)) * np.sqrt(2),
                np.flip(
                    np.arange(len(self.group_results_table.pile_tip_level_nap)) + 1
                ),
                marker="^",
                color="tab:purple",
                zorder=1,
            )
            axes.scatter(
                np.flip(data.mean(axis=0)) / np.sqrt(2),
                np.flip(
                    np.arange(len(self.group_results_table.pile_tip_level_nap)) + 1
                ),
                marker="^",
                color="tab:purple",
                zorder=1,
            )

        # Draw group result over single result
        axes.scatter(
            np.flip(self.group_results_table.__getattribute__(attribute)),
            np.flip(np.arange(len(self.group_results_table.pile_tip_level_nap)) + 1),
            marker="o",
            color="tab:red",
            zorder=1,
        )

        # Draw group result over single result
        for i, x in enumerate(data.mean(axis=0)):
            axes.annotate(f"{x.round(2)}", xy=(x, i + 1))

        # add legend to figure
        axes.legend(
            handles=[
                Patch(color=clr, label=key)
                for (key, clr) in {
                    "Single;min:max": "black",
                    "Single;Q25:Q75": "tab:blue",
                    "Single;Q50": "tab:orange",
                    "Single;mean": "tab:green",
                    "Single;mean;sqrt": "tab:purple",
                    "Group;normative": "tab:red",
                }.items()
            ],
            loc="upper left",
            bbox_to_anchor=(1, 1),
            title=f"Bandwidth {attribute}",
        )

        # set label
        axes.set_ylabel("Depth [m NAP]")
        axes.set_xlabel(f"{attribute}")

        return axes

    def plot_load_settlement(
        self,
        pile_tip_level_nap: float,
        axes: Axes | None = None,
        figsize: Tuple[float, float] = (6.0, 6.0),
        sb_max: float = 25,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot load settlement curve.

        Parameters
        ----------
        pile_tip_level_nap:
            Pile tip level  in [m NAP] for which to calculate and plot the
            load-settlement behavior.
        axes:
            Optional `Axes` object where the load-settlement data can be plotted on.
            If not provided, a new `plt.Figure` will be activated and the `Axes`
            object will be created and returned.
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        sb_max:
            Default: 25
            Maximum value for settlement at pile-tip level.
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        axes:
            The `Axes` object where the settlement curves were plotted on
        """
        # Validate required properties are present
        if self._pp.pile_type.settlement_curve is None:
            raise ValueError(
                "No settlement curve is defined for the pile-type. "
                "Please define a settlement curve in the pile-type properties."
            )

        # Validate axes
        if (axes is not None) and (not isinstance(axes, Axes)):
            raise ValueError(
                "'axes' argument to plot_load_settlement() must be a `pyplot.axes.Axes` object or None."
            )

        # Get ptl index
        idx = int(
            np.argmin(
                abs(
                    np.array(self.group_results_table.pile_tip_level_nap)
                    - pile_tip_level_nap
                )
            )
        )
        if (
            abs(self.group_results_table.pile_tip_level_nap[idx] - pile_tip_level_nap)
            > 0.01
        ):
            raise UserError(
                """No results have been calculated for the requested pile-tip-level.
                Please include this level in the pile-tip range parameter of the
                calculation."""
            )

        return get_load_settlement_plot(
            settlement_curve=self._pp.pile_type.settlement_curve,
            d_eq=self._pp.geometry.equiv_diameter_pile_tip,
            s_b=self.group_results_table.s_b[idx],
            f_c_k=self.group_results_table.F_c_k[idx],
            f_nk_k=self.group_results_table.F_nk_k[idx],
            r_b_k=self.group_results_table.R_b_k[idx],
            r_s_k=self.group_results_table.R_s_k[idx],
            rs_mob_ratio=self.group_results_table.R_s_mob_ratio[idx],
            rb_mob_ratio=self.group_results_table.R_b_mob_ratio[idx],
            sb_max=sb_max,
            axes=axes,
            figsize=figsize,
            **kwargs,
        )



from __future__ import annotations

from functools import lru_cache
from typing import Any, Dict, List, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.axes import Axes

from pypilecore.common.piles import PileProperties
from pypilecore.results.tension.single_cpt_results import SingleCPTTensionBearingResults

Number = Union[float, int]


class CPTTensionGroupResultsTable:
    """
    Dataclass that contains the bearing results of a group of CPTs.
    """

    def __init__(
        self,
        pile_tip_level_nap: Sequence[float],
        var_coef: Sequence[float],
        n_cpts: Sequence[float],
        xi_value: Sequence[float],
        xi_normative: Sequence[float],
        cpt_normative: Sequence[float],
        R_t_d_min: Sequence[float],
        R_t_d_mean: Sequence[float],
        R_t_d: Sequence[float],
        R_t_d_plug: Sequence[float],
    ):
        """
        Parameters
        ----------
        pile_tip_level_nap:
            The pile-tip level [m] w.r.t. NAP.
        var_coef:
            The variation coefficient [%] of the calculated bearing capacities in the group.
        n_cpts:
            The number of CPTs [-] that have been taken into account to establish the Xi value.
        xi_normative:
            The normative Xi (either Xi_3 or Xi_4)
        xi_value:
            The Xi value [-] that was applied to calculate the characteristic value of the
            total bearing capacity.
        cpt_normative:
            The normative CPT. Can be "group average" if that was found to be the normative scenario.
        R_t_d_min:
            The minimum of the single-CPT values for the calculated bearingcapacity [kN].
        R_t_d_mean:
            The mean of the single-CPT values for the calculated
                bearingcapacity [kN].
        R_t_d:
            calculation value of the tensile resistance of a pile or pile group
            (7.6.3.3 (a) NEN 9997-1+C2:2017) [kN]
        R_t_d_plug:
            The design value of the total plug weight bearingcapacity [kN].
        """
        self.pile_tip_level_nap = (
            np.array(pile_tip_level_nap).astype(np.float64).round(decimals=2)
        )
        """The pile-tip level [m] w.r.t. NAP."""
        self.var_coef = np.array(var_coef).astype(np.float64)
        """The variation coefficient [%] of the calculated bearing capacities in the group."""
        self.n_cpts = np.array(n_cpts).astype(np.int32)
        """The number of CPTs [-] that have been taken into account to establish the Xi value."""
        self.xi_normative = np.array(xi_normative).astype(np.str_)
        """The normative Xi (either Xi_3 or Xi_4)"""
        self.xi_value = np.array(xi_value).astype(np.float64)
        """The Xi value [-] that was applied to calculate the characteristic value of the
        total bearing capacity."""
        self.cpt_normative = np.array(cpt_normative).astype(np.str_)
        """The normative CPT. Can be "group average" if that was found to be the normative scenario."""
        self.R_t_d_min = np.array(R_t_d_min).astype(np.float64)
        """The minimum of the single-CPT values for the calculated bearingcapacity [kN]."""
        self.R_t_d_mean = np.array(R_t_d_mean).astype(np.float64)
        """The mean of the single-CPT values for the calculated
                bearingcapacity [kN]."""
        self.R_t_d = np.array(R_t_d).astype(np.float64)
        """calculation value of the tensile resistance of a pile or pile group
            (7.6.3.3 (a) NEN 9997-1+C2:2017) [kN]"""
        self.R_t_d_plug = np.array(R_t_d_plug).astype(np.float64)
        """The design value of the total plug weight bearingcapacity [kN]."""

        for value in self.__dict__.values():
            if not len(value) == len(self.pile_tip_level_nap):
                raise ValueError(
                    "Inputs for CPTGroupResults dataclass must have same length."
                )

    @lru_cache
    def to_pandas(self) -> pd.DataFrame:
        """The pandas.DataFrame representation"""
        return pd.DataFrame(self.__dict__).dropna(axis=0, how="all")

    def plot_bearing_capacities(
        self,
        axes: Axes | None = None,
        figsize: Tuple[int, int] = (8, 10),
        add_legend: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plots the design value of the bearing capacity Rcd, based on a group of CPTs.

        Parameters
        ----------
        axes:
            Optional `Axes` object where the bearing capacity data can be plotted on.
            If not provided, a new `plt.Figure` will be activated and the `Axes`
            object will be created and returned.
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        add_legend:
            Add a legend to the `Axes` object
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        axes:
            The `Axes` object where the bearing capacities were plotted on.
        """

        # Create axes objects if not provided
        if axes is not None:
            if not isinstance(axes, Axes):
                raise TypeError(
                    "'axes' argument to plot_bearing_capacities() must be a `pyplot.axes.Axes` object or None."
                )

        else:
            kwargs_subplot = {
                "figsize": figsize,
                "tight_layout": True,
            }

            kwargs_subplot.update(kwargs)

            _, axes = plt.subplots(
                1,
                1,
                **kwargs_subplot,
            )

            if not isinstance(axes, Axes):
                raise ValueError(
                    "Could not create Axes objects. This is probably due to invalid matplotlib keyword arguments. "
                )

        axes.plot(
            self.R_t_d,
            self.pile_tip_level_nap,
            "-",
            label=r"$R_{t;d}$",
        )
        axes.plot(
            self.R_t_d_plug,
            self.pile_tip_level_nap,
            ":",
            label=r"$R_{t;d;kluit}$",
        )
        axes.set_xlabel("[kN]")
        axes.set_ylabel("[m] w.r.t. NAP")

        # set grid
        axes.grid()

        # Add legend
        if add_legend:
            axes.legend(
                loc="upper left",
                bbox_to_anchor=(1, 1),
            )
        return axes


class SingleCPTTensionBearingResultsContainer:
    """A container that holds multiple SingleCPTBearingResults objects"""

    def __init__(self, cpt_results_dict: Dict[str, SingleCPTTensionBearingResults]):
        """
        Parameters
        ----------
        cpt_results_dict:
            A dictionary that maps the cpt-names to SingleCPTBearingResults objects.
        """
        self._cpt_results_dict = cpt_results_dict

    @classmethod
    def from_api_response(
        cls, cpt_results_list: list, cpt_input: dict
    ) -> "SingleCPTTensionBearingResultsContainer":
        """
        Instantiates the SingleCPTBearingResultsContainer object from the "cpts" array,
        which is returned in the response of a "compression/multiple-cpts/results" endpoint call.
        """
        return cls(
            cpt_results_dict={
                cpt_results[
                    "test_id"
                ]: SingleCPTTensionBearingResults.from_api_response(
                    cpt_results_dict=cpt_results,
                    ref_height=cpt_input[cpt_results["test_id"]]["ref_height"],
                    surface_level_ref=cpt_input[cpt_results["test_id"]][
                        "surface_level_nap"
                    ],
                    x=cpt_input[cpt_results["test_id"]].get("location", {}).get("x"),
                    y=cpt_input[cpt_results["test_id"]].get("location", {}).get("y"),
                )
                for cpt_results in cpt_results_list
            }
        )

    def __getitem__(self, test_id: str) -> SingleCPTTensionBearingResults:
        if not isinstance(test_id, str):
            raise TypeError(f"Expected a test-id as a string, but got: {type(test_id)}")

        return self.get_cpt_results(test_id)

    @property
    def cpt_results_dict(self) -> Dict[str, SingleCPTTensionBearingResults]:
        """The dictionary that maps the cpt-names to SingleCPTBearingResults objects."""
        return self._cpt_results_dict

    @property
    def test_ids(self) -> List[str]:
        """The test-ids of the CPTs."""
        return list(self.cpt_results_dict.keys())

    @property
    def results(self) -> List[SingleCPTTensionBearingResults]:
        """The computed results, as a list of SingleCPTBearingResults objects."""
        return list(self.cpt_results_dict.values())

    def get_cpt_results(self, test_id: str) -> SingleCPTTensionBearingResults:
        """
        Returns the `SingleCPTBearingResults` object for the provided test_id.
        """

        if test_id not in self.cpt_results_dict.keys():
            raise ValueError(
                f"No Cpt-results were calculated for this test-id: {test_id}. "
                "Please check the spelling or run a new calculation for this CPT."
            )

        return self.cpt_results_dict[test_id]

    def get_results_per_cpt(self, column_name: str) -> pd.DataFrame:
        """
        Returns a pandas dataframe with a single result-item, organized per CPT
        (test-id) and pile-tip-level-nap.

        Parameters
        ----------
        column_name:
            The name of the result-item / column name of the single-cpt-results table.
        """
        if column_name not in self.to_pandas().columns or column_name in [
            "pile_tip_level_nap",
            "test_id",
        ]:
            raise ValueError(f"Invalid column_name: {column_name} provided.")

        results = pd.pivot_table(
            self.to_pandas(),
            values=column_name,
            index="pile_tip_level_nap",
            columns="test_id",
            dropna=False,
        )
        return results.sort_values("pile_tip_level_nap", ascending=False)

    @lru_cache
    def to_pandas(self) -> pd.DataFrame:
        """Returns a total overview of all single-cpt results in a pandas.DataFrame representation."""
        df_list: List[pd.DataFrame] = []

        for test_id in self.cpt_results_dict:
            df = self.cpt_results_dict[test_id].table.to_pandas()
            df = df.assign(test_id=test_id)
            df_list.append(df)

        cpt_results_df = pd.concat(df_list)
        cpt_results_df = cpt_results_df.assign(
            pile_tip_level_nap=cpt_results_df.pile_tip_level_nap.round(2)
        )

        return cpt_results_df


class MultiCPTTensionBearingResults:
    """
    Object that contains the results of a PileCore multi-cpt calculation.

    *Not meant to be instantiated by the user.*
    """

    def __init__(
        self,
        cpt_results: SingleCPTTensionBearingResultsContainer,
        group_results_table: CPTTensionGroupResultsTable,
        pile_properties: PileProperties,
        gamma_s_t: float,
        gamma_m_var_qc: float,
        soil_load: float,
    ) -> None:
        """
        Parameters
        ----------
        cpt_results:
            The container object with single-CPT results
        group_results_table:
            The table object with CPT-group-results
        pile_properties:
            The PileProperties object
        gamma_s_t:
            Safety factor γ_s_t. Used to determine the tensile resistance of a pile. [-]
        gamma_m_var_qc:
            partial factor for the influence of load switching; NEN 9997-1+C2:2017 7.6.3.3 (d). [-]
        phi_plug:
            The effective angle of internal friction of the soil layer based on pile type [degrees].
        soil_load:
            (Fictive) load on soil used to calculate soil settlement [kPa]. This is
            required and used to determine settlement of pile w.r.t. soil.
        """
        self._pp = pile_properties
        self._gamma_s_t = gamma_s_t
        self._gamma_m_var_qc = gamma_m_var_qc
        self._soil_load = soil_load

        self._cpt_results = cpt_results

        self._group_results_table = group_results_table

    @classmethod
    def from_api_response(
        cls, response_dict: dict, cpt_input: dict
    ) -> "MultiCPTTensionBearingResults":
        """
        Build the object from the response payload of the PileCore endpoint
        "/compression/multi-cpt/results".
        """
        cpt_results_dict = SingleCPTTensionBearingResultsContainer.from_api_response(
            cpt_results_list=response_dict["cpts"], cpt_input=cpt_input
        )
        group_results = response_dict["group_results"]
        return cls(
            cpt_results=cpt_results_dict,
            pile_properties=PileProperties.from_api_response(
                response_dict["pile_properties"]
            ),
            group_results_table=CPTTensionGroupResultsTable(
                pile_tip_level_nap=group_results["pile_tip_level_nap"],
                var_coef=group_results["var_coef"],
                n_cpts=group_results["n_cpts"],
                xi_value=group_results["xi_value"],
                xi_normative=group_results["xi_normative"],
                cpt_normative=group_results["cpt_normative"],
                R_t_d_min=group_results["R_t_d_min"],
                R_t_d_mean=group_results["R_t_d_mean"],
                R_t_d=group_results["R_t_d"],
                R_t_d_plug=group_results["R_t_d_plug"],
            ),
            gamma_s_t=response_dict["calculation_params"]["gamma_s_t"],
            soil_load=response_dict["calculation_params"]["soil_load"],
            gamma_m_var_qc=response_dict["calculation_params"]["gamma_m_var_qc"],
        )

    @property
    def pile_properties(self) -> PileProperties:
        """
        The PileProperties object.
        """
        return self._pp

    @property
    def cpt_results(self) -> SingleCPTTensionBearingResultsContainer:
        """The SingleCPTBearingResultsContainer object."""
        return self._cpt_results

    @property
    def cpt_names(self) -> List[str]:
        """The test-ids of the CPTs."""
        return self.cpt_results.test_ids

    @property
    def group_results_table(self) -> CPTTensionGroupResultsTable:
        """The CPTGroupResultsTable dataclass, containing the group results."""
        return self._group_results_table



from typing import Any, List, Union

from matplotlib.axes import Axes


def validate_axes_array(axes: Any, shape: Union[int, List[int]]) -> None:
    """
    Validates if the axes argument is an iterable object of a certain shape, filled with
    Axes objects.

    Parameters
    ----------
    axes:
        The presumed array of `Axes` objects
    shape:
        The shape of the `axes` array.
    """
    if isinstance(shape, int):
        shape = [shape]

    try:
        assert len(axes) == shape[0]

        for idx in range(shape[0]):
            if len(shape) > 1:
                validate_axes_array(axes[idx], shape[1:])

            else:
                assert axes[idx] is None or isinstance(axes[idx], Axes)

    except (IndexError, AssertionError, ValueError, TypeError):
        raise ValueError(
            f"""Invalid value for axes argument.
        Provide an array-like with Optional[Axes] objects and shape {shape}."""
        )



from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Dict, List, Sequence, Tuple

import natsort
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from numpy.typing import NDArray
from shapely import MultiPoint

from pypilecore.results.compression.multi_cpt_results import (
    MultiCPTCompressionBearingResults,
)
from pypilecore.results.post_processing import (
    MaxBearingResult,
    MaxBearingResults,
    MaxBearingTable,
)


class SingleClusterData:
    """
    Data for a single CPT subgroup
    *Not meant to be instantiated by the user.*
    """

    def __init__(
        self,
        characteristic_bearing_capacity: Sequence[float],
        design_bearing_capacity: Sequence[float],
        design_negative_friction: Sequence[float],
        group_centre_to_centre_validation: Sequence[bool],
        group_centre_to_centre_validation_15: Sequence[bool],
        group_centre_to_centre_validation_20: Sequence[bool],
        group_centre_to_centre_validation_25: Sequence[bool],
        mean_calculated_bearing_capacity: Sequence[float],
        min_calculated_bearing_capacity: Sequence[float],
        net_design_bearing_capacity: Sequence[float],
        nominal_cpt: Sequence[str],
        pile_tip_level: Sequence[float],
        variation_coefficient: Sequence[float],
        xi_factor: Sequence[str],
        xi_values: Sequence[float],
    ):
        """
        Parameters
        ----------
        characteristic_bearing_capacity:
            characteristic bearing capacity [kN]
        design_bearing_capacity:
            design bearing capacity [kN]
        design_negative_friction:
            design negative friction [kN]
        group_centre_to_centre_validation:
            group centre to centre validation
        group_centre_to_centre_validation_15:
            group centre to centre validation 15 meter
        group_centre_to_centre_validation_20:
            group centre to centre validation 20 meter
        group_centre_to_centre_validation_25:
            group centre to centre validation 25 meter
        mean_calculated_bearing_capacity:
            mean calculated bearing capacity [kN]
        min_calculated_bearing_capacity:
            min calculated bearing capacity [kN]
        net_design_bearing_capacity:
            net design bearing capacity [kN]
        nominal_cpt:
            nominal cpt
        pile_tip_level:
            pile tip level [m w.r.t NAP]
        variation_coefficient:
            variation coefficient [-]
        xi_factor:
            xi factor
        xi_values:
            xi values [-]
        """
        self._characteristic_bearing_capacity = characteristic_bearing_capacity
        self._design_bearing_capacity = design_bearing_capacity
        self._design_negative_friction = design_negative_friction
        self._group_centre_to_centre_validation = group_centre_to_centre_validation
        self._group_centre_to_centre_validation_15 = (
            group_centre_to_centre_validation_15
        )
        self._group_centre_to_centre_validation_20 = (
            group_centre_to_centre_validation_20
        )
        self._group_centre_to_centre_validation_25 = (
            group_centre_to_centre_validation_25
        )
        self._mean_calculated_bearing_capacity = mean_calculated_bearing_capacity
        self._min_calculated_bearing_capacity = min_calculated_bearing_capacity
        self._net_design_bearing_capacity = net_design_bearing_capacity
        self._nominal_cpt = nominal_cpt
        self._pile_tip_level = pile_tip_level
        self._variation_coefficient = variation_coefficient
        self._xi_factor = xi_factor
        self._xi_values = xi_values

        raw_lengths = [len(values) for values in self.__dict__.values()]
        if len(list(set(raw_lengths))) > 1:
            raise ValueError("All values in this dataclass must have the same length.")

    @property
    def characteristic_bearing_capacity(self) -> NDArray[np.float64]:
        """Characteristic bearing capacity [kN]"""
        return np.array(self._characteristic_bearing_capacity).astype(np.float64)

    @property
    def design_bearing_capacity(self) -> NDArray[np.float64]:
        """Design bearing capacity [kN]"""
        return np.array(self._design_bearing_capacity).astype(np.float64)

    @property
    def design_negative_friction(self) -> NDArray[np.float64]:
        """Design negative friction [kN]"""
        return np.array(self._design_negative_friction).astype(np.float64)

    @property
    def group_centre_to_centre_validation(self) -> NDArray[np.bool_]:
        """Group centre to centre validation"""
        return np.array(self._group_centre_to_centre_validation).astype(np.bool_)

    @property
    def group_centre_to_centre_validation_15(self) -> NDArray[np.bool_]:
        """Group centre to centre validation 15 meter"""
        return np.array(self._group_centre_to_centre_validation_15).astype(np.bool_)

    @property
    def group_centre_to_centre_validation_20(self) -> NDArray[np.bool_]:
        """Group centre to centre validation 20 meter"""
        return np.array(self._group_centre_to_centre_validation_20).astype(np.bool_)

    @property
    def group_centre_to_centre_validation_25(self) -> NDArray[np.bool_]:
        """Group centre to centre validation 25 meter"""
        return np.array(self._group_centre_to_centre_validation_25).astype(np.bool_)

    @property
    def mean_calculated_bearing_capacity(self) -> NDArray[np.float64]:
        """Mean calculated bearing capacity [kN]"""
        return np.array(self._mean_calculated_bearing_capacity).astype(np.float64)

    @property
    def min_calculated_bearing_capacity(self) -> NDArray[np.float64]:
        """Min calculated bearing capacity [kN]"""
        return np.array(self._min_calculated_bearing_capacity).astype(np.float64)

    @property
    def net_design_bearing_capacity(self) -> NDArray[np.float64]:
        """Net design bearing capacity [kN]"""
        return np.array(self._net_design_bearing_capacity).astype(np.float64)

    @property
    def nominal_cpt(self) -> NDArray[np.str_]:
        """Nominal cpt"""
        return np.array(self._nominal_cpt).astype(str)

    @property
    def pile_tip_level(self) -> NDArray[np.float64]:
        """Pile tip level [m w.r.t NAP]"""
        return np.array(self._pile_tip_level).astype(np.float64)

    @property
    def variation_coefficient(self) -> NDArray[np.float64]:
        """Variation coefficient [-]"""
        return np.array(self._variation_coefficient).astype(np.float64)

    @property
    def xi_factor(self) -> NDArray[np.str_]:
        """Xi factor"""
        return np.array(self._xi_factor).astype(str)

    @property
    def xi_values(self) -> NDArray[np.float64]:
        """Xi values [-]"""
        return np.array(self._xi_values).astype(np.float64)

    @cached_property
    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame(self.__dict__)

    def plot_variation_coefficient(
        self, axes: Axes | None = None, **kwargs: Any
    ) -> None:
        """
        Plot the bearing capacity and variation coefficient in a subplot

        Parameters
        ----------
        axes:
            `plt.Axes` object where the data can be plotted on.
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.
        """
        if axes is None:
            _, axes = plt.subplots(**kwargs)

        # create variation coefficient plot
        axes.plot(self.variation_coefficient, self.pile_tip_level, "o-")
        axes.axvline(x=0.12, color="black", linestyle="--")
        axes.grid()
        axes.set_xlabel("Variation coefficient [-]")

    def plot_bearing_capacity(
        self, axes: Axes | None = None, pile_load_uls: float = 0.0, **kwargs: Any
    ) -> None:
        """
        Plot the bearing capacity and variation coefficient in a subplot

        Note
        ------
        For the `Net bearing capacity` subplot there are two colors plotted:
         - orange:  conservative bearing capacity
         - blue:    net bearing capacity

        Parameters
        ----------
        axes:
            `plt.Axes` object where the data can be plotted on.
        pile_load_uls:
            Default is 0.0
            ULS load in kN.
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.
        """
        if axes is None:
            _, axes = plt.subplots(**kwargs)

        # add net bearing capacity to plot
        axes.scatter(
            self.net_design_bearing_capacity,
            self.pile_tip_level,
            color=list(
                map(
                    lambda x: "tab:blue" if x <= 0.12 else "tab:orange",
                    self.variation_coefficient,
                )
            ),
        )
        axes.axvline(x=pile_load_uls, color="black", linestyle="--")
        axes.grid()
        axes.set_xlabel("Net bearing capacity [kN]")

    def plot_group_centre_to_centre_validation(
        self, axes: plt.Axes | None = None, **kwargs: Any
    ) -> None:
        """
        Plot the spacing checks in a subplot

        Note
        ------
        For the `spacing` subplot there are two colors plotted:
         - red:     invalid spacing
         - green:   valid spacing

        Parameters
        ----------
        axes:
            `plt.Axes` object where the data can be plotted on.
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.
        """
        if axes is None:
            _, axes = plt.subplots(**kwargs)

        axes.scatter(
            [0] * len(self.pile_tip_level),
            self.pile_tip_level,
            marker="o",
            color=list(
                map(
                    lambda x: "tab:green" if x else "tab:red",
                    self.group_centre_to_centre_validation_25,
                )
            ),
        )
        axes.scatter(
            [1] * len(self.pile_tip_level),
            self.pile_tip_level,
            marker="s",
            color=list(
                map(
                    lambda x: "tab:green" if x else "tab:red",
                    self.group_centre_to_centre_validation_20,
                )
            ),
        )
        axes.scatter(
            [2] * len(self.pile_tip_level),
            self.pile_tip_level,
            marker="D",
            color=list(
                map(
                    lambda x: "tab:green" if x else "tab:red",
                    self.group_centre_to_centre_validation_15,
                )
            ),
        )
        axes.grid()
        axes.set_xticks([0, 1, 2], ["25", "20", "15"])
        axes.set_xlabel("CPT ctc [m]")

    def plot_xi(self, axes: plt.Axes | None = None, **kwargs: Any) -> None:
        """
        Plot the xi factor in a subplot

        Note
        ------
        For the `xi factor` subplot there are two colors plotted:
         - olive:   xi3
         - cyan:    xi4

        Parameters
        ----------
        axes:
            `plt.Axes` object where the data can be plotted on.
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.
        """
        if axes is None:
            _, axes = plt.subplots(**kwargs)

        axes.scatter(
            self.xi_values,
            self.pile_tip_level,
            color=list(
                map(
                    lambda i: "tab:cyan" if i == "\u03BE4" else "tab:olive",
                    self.xi_factor,
                )
            ),
        )
        axes.grid()
        axes.set_xlabel("xi value [-]")


@dataclass(frozen=True)
class SingleClusterResult:
    """
    *Not meant to be instantiated by the user.*

    Attributes:
    ------------
    cpt_names: List[str]
        List of cpt names present in this cluster
    coordinates: List[Tuple[float, float]]
        List of coordinates present in this cluster
    pile_load_uls
        ULS load in kN. Used to determine if a grouping configuration is valid.
    maximum_pile_level: float
        maximum pile level [m w.r.t NAP]
    minimum_pile_level: float
        minimum pile level [m w.r.t NAP]
    number_of_consecutive_pile_levels: int
        number of consecutive pile levels
    pile_load_check: bool
        True if a minimum design pile bearing capacity based on the given pile load ULS at one or more
        pile-tip levels.
    spatial_check: bool
        True if cluster is spatially coherent, which means there are no other CPTs in between the members
        of the subgroup.
    variation_check: bool
        True if a maximum variation coefficient of 12% at one or more pile-tip levels.
    centre_to_centre_check: bool
        True if one of the conditions stated in NEN9997-1 3.2.3 is met at one or more pile-tip levels.
    data: SingleClusterData
        single cluster dataclass
    """

    cpt_names: List[str]
    coordinates: List[Tuple[float, float]]
    pile_load_uls: float
    maximum_pile_level: float
    minimum_pile_level: float
    number_of_consecutive_pile_levels: int
    pile_load_check: bool
    spatial_check: bool
    variation_check: bool
    centre_to_centre_check: bool
    data: SingleClusterData

    @classmethod
    def from_api_response(
        cls, response_dict: dict, pile_load_uls: float
    ) -> "SingleClusterResult":
        try:
            table = response_dict["table"]
            return cls(
                cpt_names=response_dict["names"],
                coordinates=response_dict["coordinates"],
                pile_load_uls=pile_load_uls,
                maximum_pile_level=response_dict["maximum_pile_level"],
                minimum_pile_level=response_dict["minimum_pile_level"],
                number_of_consecutive_pile_levels=response_dict[
                    "number_of_consecutive_pile_levels"
                ],
                pile_load_check=response_dict["pile_load_check"],
                spatial_check=response_dict["spatial_check"],
                variation_check=response_dict["variation_check"],
                centre_to_centre_check=response_dict["centre_to_centre_check"],
                data=SingleClusterData(
                    characteristic_bearing_capacity=table[
                        "characteristic_bearing_capacity"
                    ],
                    design_bearing_capacity=table["design_bearing_capacity"],
                    design_negative_friction=table["design_negative_friction"],
                    group_centre_to_centre_validation=table[
                        "group_centre_to_centre_validation"
                    ],
                    group_centre_to_centre_validation_15=table[
                        "group_centre_to_centre_validation_15"
                    ],
                    group_centre_to_centre_validation_20=table[
                        "group_centre_to_centre_validation_20"
                    ],
                    group_centre_to_centre_validation_25=table[
                        "group_centre_to_centre_validation_25"
                    ],
                    mean_calculated_bearing_capacity=table[
                        "mean_calculated_bearing_capacity"
                    ],
                    min_calculated_bearing_capacity=table[
                        "min_calculated_bearing_capacity"
                    ],
                    net_design_bearing_capacity=table["net_design_bearing_capacity"],
                    nominal_cpt=table["nominal_cpt"],
                    pile_tip_level=[round(elem, 2) for elem in table["pile_tip_level"]],
                    variation_coefficient=table["variation_coefficient"],
                    xi_factor=table["xi_factor"],
                    xi_values=table["xi_values"],
                ),
            )
        except KeyError as e:
            raise KeyError(
                "Response dictionary is missing an expected key.\n" rf"Traceback: {e}"
            )
        except ValueError as e:
            raise ValueError(
                f"Could not create `SingleClusterResult` class with the following cpts: {response_dict.get('names')} \n"
                rf"Traceback: {e}"
            )

    def map(
        self,
        figsize: Tuple[int, int] | None = None,
        **kwargs: Any,
    ) -> plt.Figure:
        """
        Plot a map of the cpt locations

        Parameters
        ----------
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        axes:
            The `Axes` object where the data was plotted on.
        """

        # Create axes objects if not provided
        kwargs_subplot = {
            "figsize": figsize,
            "tight_layout": True,
        }

        kwargs_subplot.update(kwargs)

        fig, axes = plt.subplots(
            **kwargs_subplot,
        )

        # plot cpt
        xy = list(zip(*self.coordinates))
        axes = axes.scatter(x=xy[0], y=xy[1], color="grey")

        # add labels (cpt names) to map
        for x, y, label in zip(xy[0], xy[1], self.cpt_names):
            axes.annotate(label, xy=(x, y), xytext=(3, 3), textcoords="offset points")
        axes.ticklabel_format(useOffset=False, style="plain")
        return fig

    def plot(
        self,
        figsize: Tuple[int, int] | None = None,
        **kwargs: Any,
    ) -> plt.Figure:
        """
        Plot contains the:
            - bearing capacity
            - variation coefficient
            - xi factor
            - centre to centre validation

        Note
        ------
        For the `Net bearing capacity` subplot there are two colors plotted:
         - orange:  conservative bearing capacity
         - blue:    net bearing capacity

        For the `xi factor` subplot there are two colors plotted:
         - olive:   xi3
         - cyan:    xi4

        For the `spacing` subplot there are two colors plotted:
         - red:     invalid spacing
         - green:   valid spacing

        Parameters
        ----------
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        figure:
            `plt.Figure` object.
        """

        kwargs_subplot = {
            "sharey": "row",
            "sharex": "col",
            "figsize": figsize,
        }

        kwargs_subplot.update(kwargs)

        fig, axes = plt.subplots(
            1,
            4,
            **kwargs_subplot,
        )
        # add plot variation coefficient
        self.data.plot_variation_coefficient(axes[0])

        # add plot bearing capacity
        self.data.plot_bearing_capacity(axes[1], self.pile_load_uls)

        # add xi table
        self.data.plot_xi(axes[2])

        # add centre to centre
        self.data.plot_group_centre_to_centre_validation(axes[3])

        return fig


@dataclass(frozen=True)
class GrouperResults:
    """
    *Not meant to be instantiated by the user.*

    Use the `from_api_response` method to instantiate the class.

    Attributes:
    ------------
    clusters: List[SingleClusterResult]
    """

    clusters: List[SingleClusterResult]
    multi_cpt_bearing_results: MultiCPTCompressionBearingResults

    def __post_init__(self) -> None:
        for cluster in self.clusters:
            for cpt_name in cluster.cpt_names:
                # check if the cpt names in the SingleClusterResults are also present
                # in the MultiCPTBearingResults
                if (
                    cpt_name
                    not in self.multi_cpt_bearing_results.cpt_results.cpt_results_dict.keys()
                ):
                    raise ValueError(
                        "CPT names dont match between MultiCPTBearingResults object and GrouperResults. "
                        "Make sure that you use the same MultiCPTBearingResults as you generated "
                        "the subgroups/clusters with."
                    )

                # Check that all the pile tip levels in the SingleClusterResults are
                # also present in the MultiCPTBearingResults
                for pile_tip_level in cluster.data.pile_tip_level:
                    if not np.isclose(
                        pile_tip_level,
                        self.multi_cpt_bearing_results.cpt_results.cpt_results_dict[
                            cpt_name
                        ].table.pile_tip_level_nap,
                        rtol=1e-2,
                    ).any():
                        raise ValueError(
                            "Pile tip levels dont match between MultiCPTBearingResults object and GrouperResults. "
                            "Make sure that you use the same MultiCPTBearingResults as you generated "
                            "the subgroups/clusters with."
                        )

    @classmethod
    def from_api_response(
        cls,
        response_dict: dict,
        pile_load_uls: float,
        multi_cpt_bearing_results: MultiCPTCompressionBearingResults,
    ) -> "GrouperResults":
        """
        Stores the response of the PileCore endpoint
        "/grouper/group_cpts"

        Parameters
        ----------
        response_dict:
           The resulting response of a call to `get_groups_api_result()`
        pile_load_uls:
            ULS load in kN. Used to determine if a grouping configuration is valid.
        multi_cpt_bearing_results:
           The container that holds multiple SingleCPTBearingResults objects
        """
        results = [
            SingleClusterResult.from_api_response(item, pile_load_uls)
            for item in response_dict["sub_groups"]
        ]
        return cls(
            clusters=results, multi_cpt_bearing_results=multi_cpt_bearing_results
        )

    @cached_property
    def max_bearing_results(self) -> "MaxBearingResults":
        """
        Get the results of the maximum net design bearing capacity (R_c_d_net) for every CPT.
        """
        max_bearing: Dict[str, Any] = {}

        # iterate over single cpt result
        for (
            cpt_name,
            _single_cpt_result,
        ) in self.multi_cpt_bearing_results.cpt_results.cpt_results_dict.items():
            single_cpt_result = deepcopy(_single_cpt_result)
            max_bearing[cpt_name] = dict(
                pile_head_level_nap=single_cpt_result.pile_head_level_nap,
                soil_properties=single_cpt_result.soil_properties,
                results_table=dict(
                    pile_tip_level_nap=single_cpt_result.table.pile_tip_level_nap,
                    R_c_d_net=single_cpt_result.table.R_c_d_net,
                    F_nk_d=single_cpt_result.table.F_nk_d,
                    origin=[f"CPT:{cpt_name}"]
                    * len(single_cpt_result.table.pile_tip_level_nap),
                ),
            )

        # iterate over subgroups result
        for cluster_idx, cluster in enumerate(self.clusters):
            # iterate over cpts in subgroup
            for cpt_name in cluster.cpt_names:
                # iterate over pile tip levels in the cluster results for the cpt
                for cluster_ptl_idx, ptl in enumerate(cluster.data.pile_tip_level):
                    # find corresponding pile tip level index in the max_bearing results
                    max_bearing_ptl_idx = np.abs(
                        max_bearing[cpt_name]["results_table"]["pile_tip_level_nap"]
                        - ptl
                    ).argmin()

                    # check bearing capacity
                    if cluster.data.net_design_bearing_capacity[
                        cluster_ptl_idx
                    ] > np.nan_to_num(
                        max_bearing[cpt_name]["results_table"]["R_c_d_net"][
                            max_bearing_ptl_idx
                        ]
                    ):
                        # replace data
                        max_bearing[cpt_name]["results_table"]["R_c_d_net"][
                            max_bearing_ptl_idx
                        ] = cluster.data.net_design_bearing_capacity[cluster_ptl_idx]
                        max_bearing[cpt_name]["results_table"]["F_nk_d"][
                            max_bearing_ptl_idx
                        ] = cluster.data.design_negative_friction[cluster_ptl_idx]
                        max_bearing[cpt_name]["results_table"]["origin"][
                            max_bearing_ptl_idx
                        ] = f"Group:{cluster_idx}"

        return MaxBearingResults(
            cpt_results_dict={
                cpt_name: MaxBearingResult(
                    pile_head_level_nap=data["pile_head_level_nap"],
                    soil_properties=data["soil_properties"],
                    table=MaxBearingTable(
                        pile_tip_level_nap=data["results_table"]["pile_tip_level_nap"],
                        R_c_d_net=data["results_table"]["R_c_d_net"],
                        F_nk_d=data["results_table"]["F_nk_d"],
                        origin=data["results_table"]["origin"],
                    ),
                )
                for cpt_name, data in max_bearing.items()
            }
        )

    def map(
        self,
        distance: float = 25.0,
        add_tags: bool = True,
        figsize: Tuple[int, int] | None = None,
        **kwargs: Any,
    ) -> plt.Figure:
        """
        Plot a map of the valid subgroups.
        Plot contains the:

            - convex_hull of the subgroup with buffer distance
            - All CPT's with tag

        Parameters
        ----------
        distance : float, optional
            Default is 25.
            The buffer around the convex_hull of the subgroup
        add_tags : bool, optional
            default is True
            Show the CTP names as tags on the map
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        figure:
            The `Figure` object where the data was plotted on.
        """

        kwargs_subplot = {
            "figsize": figsize,
            "tight_layout": True,
        }

        kwargs_subplot.update(kwargs)

        figure, axes = plt.subplots(
            **kwargs_subplot,
        )

        for group_id, cluster in enumerate(self.clusters):
            # add cpts to plot
            xy = list(zip(*cluster.coordinates))
            axes.scatter(xy[0], xy[1], color="grey")

            if add_tags:
                for x, y, label in zip(xy[0], xy[1], cluster.cpt_names):
                    axes.annotate(
                        label, xy=(x, y), xytext=(3, 3), textcoords="offset points"
                    )

            # add group convex hull
            polygon = (
                MultiPoint(cluster.coordinates)
                .convex_hull.buffer(distance=distance)
                .exterior
            )
            axes.plot(polygon.xy[0], polygon.xy[1], label=f"Group {group_id}")

        axes.legend(bbox_to_anchor=(1, 1), loc="upper left")
        axes.ticklabel_format(useOffset=False, style="plain")

        return figure

    def plot(
        self,
        figsize: Tuple[int, int] | None = None,
        **kwargs: Any,
    ) -> plt.Figure:
        """
        Plot a summary of the valid subgroups.

        Note
        -----
        Plot contains the:

            - cpts within a subgroup
                - green:
                    There are no other CPTs in between the members of the subgroup. The
                    group is also compliant with the NEN9997-1 3.2.3 centre to centre
                    validation.
                - orange:
                    There are no other CPTs in between the members of the subgroup. The
                    centre-to-centre check failed and so the group does not follow the
                    NEN9997-1 3.2.3 centre to centre validation.
                - red:
                    There are other CPTs in between the members of the subgroup.
            - valid depth of the subgroup

        Parameters
        ----------
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        figure:
            the `plt.Figure` object.
        """
        kwargs_subplot = {
            "sharex": "row",
            "figsize": figsize,
            "tight_layout": True,
        }

        kwargs_subplot.update(kwargs)

        figure, axes = plt.subplots(
            1,
            2,
            **kwargs_subplot,
        )

        # place holds needed to sort for plot
        group_id_list_sort = []
        cpt_names_list: List[str] = []
        color_list = []
        group_id_list = []

        for group_id, cluster in enumerate(self.clusters):
            group_id_list_sort.extend([group_id] * len(cluster.cpt_names))
            cpt_names_list.extend(cluster.cpt_names)
            color_list.extend(
                [
                    "tab:green"
                    if cluster.spatial_check and cluster.centre_to_centre_check
                    else "tab:orange"
                    if cluster.spatial_check and ~cluster.centre_to_centre_check
                    else "tab:red"
                ]
                * len(cluster.cpt_names)
            )
            valid_pile_level = np.array(cluster.data.pile_tip_level)[
                np.array(cluster.data.net_design_bearing_capacity)
                >= cluster.pile_load_uls
            ]
            axes[1].scatter([group_id] * len(valid_pile_level), valid_pile_level)
            group_id_list.append(group_id)

            data = pd.DataFrame(
                {
                    "group_id": group_id_list_sort,
                    "cpt_names": cpt_names_list,
                    "colors": color_list,
                }
            ).sort_values("cpt_names", ascending=False, key=natsort.natsort_keygen())
            axes[0].scatter(x="group_id", y="cpt_names", color="colors", data=data)
            axes[0].set_xlabel("Group ID")
            axes[0].set_xticks(group_id_list)
            axes[0].set_ylabel("CPT name")
            axes[0].grid(which="major", axis="both", alpha=0.1)

        axes[1].set_xlabel("Group ID")
        axes[1].set_ylabel("Depth [m NAP]")
        axes[1].grid(which="major", axis="both", alpha=0.1)

        return figure



import logging
from time import sleep
from typing import Literal

from nuclei.client import NucleiClient
from requests import Response


def wait_until_ticket_is_ready(
    client: NucleiClient, ticket: Response, verbose: bool = False
) -> None:
    if verbose:
        logging.info("Waiting for ticket to be ready")

    if ticket.status_code != 200:
        raise RuntimeError(rf"{ticket.text}")

    if verbose:
        logging.info("Ticket status: OK")
        logging.info(f"Ticket ID: {ticket.json()['id']}")

    status = "STARTED"
    sleep_time = 0.05

    while status in ["PENDING", "STARTED", "RETRY"]:
        # Exponential backoff
        sleep_time = min(sleep_time * 2, 10)
        if verbose:
            logging.info("Sleeping for %s seconds", sleep_time)
        sleep(sleep_time)

        # Get the status of the ticket
        if verbose:
            logging.info("Polling ticket status")
        status_response = client.call_endpoint(
            "PileCore",
            "/get-task-status",
            version="v3",
            schema=ticket.json(),
            return_response=True,
        )

        # Check if the status response is OK
        if status_response.status_code != 200:
            raise RuntimeError(rf"{status_response.text}")

        status = status_response.json()["state"]
        if verbose:
            logging.info("Ticket status: %s", status)

    # If the status is FAILURE, raise an error
    if status == "FAILURE":
        # Get the task-status failure message
        failure_message = status_response.json()["msg"]

        # Try to get the task-result failure message
        try:
            result_response = client.call_endpoint(
                "PileCore",
                "/get-task-result",
                version="v3",
                schema=ticket.json(),
                return_response=True,
            )
            failure_message = result_response.text

        # Raise the obtained failure message
        finally:
            raise RuntimeError(failure_message)


def get_multi_cpt_api_result(
    client: NucleiClient, payload: dict, verbose: bool = False
) -> dict:
    """
    Wrapper around the PileCore endpoint "/compression/multiple-cpts/results".

    Parameters
    ----------
    client: NucleiClient
        client object created by [nuclei](https://github.com/cemsbv/nuclei)
    payload: dict
        the payload of the request, can be created by calling `create_grouper_payload()`
    verbose: bool
        if True, print additional information to the console
    """
    logging.info(
        "Calculating bearing capacities... \n"
        "Depending on the amount of pile tip levels and CPT's this can take a while."
    )
    ticket = client.call_endpoint(
        "PileCore",
        "/compression/multiple-cpts/results",
        version="v3",
        schema=payload,
        return_response=True,
    )

    wait_until_ticket_is_ready(client=client, ticket=ticket, verbose=verbose)

    return client.call_endpoint(
        "PileCore", "/get-task-result", version="v3", schema=ticket.json()
    )


def get_multi_cpt_api_report(
    client: NucleiClient, payload: dict, verbose: bool = False
) -> dict:
    """
    Wrapper around the PileCore endpoint "/compression/multiple-cpts/report".

    Parameters
    ----------
    client: NucleiClient
        client object created by [nuclei](https://github.com/cemsbv/nuclei)
    payload: dict
        the payload of the request, can be created by calling `create_grouper_payload()`
    verbose: bool
        if True, print additional information to the console
    """
    logging.info(
        "Generate report... \n"
        "Depending on the amount of pile tip levels and CPT's this can take a while."
    )
    ticket = client.call_endpoint(
        "PileCore",
        "/compression/multiple-cpts/report",
        version="v3",
        schema=payload,
        return_response=True,
    )
    wait_until_ticket_is_ready(client=client, ticket=ticket, verbose=verbose)

    return client.call_endpoint(
        "PileCore", "/get-task-result", version="v3", schema=ticket.json()
    )


def get_groups_api_result(
    client: NucleiClient, payload: dict, verbose: bool = False
) -> dict:
    """
    Wrapper around the PileCore endpoint "/grouper/group_cpts".

    Parameters
    ----------
    client: NucleiClient
        client object created by [nuclei](https://github.com/cemsbv/nuclei)
    payload: dict
        the payload of the request, can be created by calling `create_grouper_payload()`
    verbose: bool
        if True, print additional information to the console
    """
    logging.info(
        "Finding groups... \n"
        "Depending on the amount of pile tip levels and CPT's this can take a while."
    )
    ticket = client.call_endpoint(
        "PileCore",
        "/grouper/group_cpts",
        version="v3",
        schema=payload,
        return_response=True,
    )

    wait_until_ticket_is_ready(client=client, ticket=ticket, verbose=verbose)

    return client.call_endpoint(
        "PileCore", "/get-task-result", version="v3", schema=ticket.json()
    )


def get_optimize_groups_api_result(
    client: NucleiClient, payload: dict, verbose: bool = False
) -> dict:
    """
    Wrapper around the PileCore endpoint "/grouper/optimize_groups".

    Parameters
    ----------
    client: NucleiClient
        client object created by [nuclei](https://github.com/cemsbv/nuclei)
    payload: dict
        the payload of the request, can be created by calling `create_grouper_payload()`
    verbose: bool
        if True, print additional information to the console
    """
    logging.info(
        "Optimize groups... \n"
        "Depending on the amount of pile tip levels and CPT's this can take a while."
    )
    ticket = client.call_endpoint(
        "PileCore",
        "/grouper/optimize_groups",
        version="v3",
        schema=payload,
        return_response=True,
    )

    wait_until_ticket_is_ready(client=client, ticket=ticket, verbose=verbose)

    return client.call_endpoint(
        "PileCore", "/get-task-result", version="v3", schema=ticket.json()
    )


def get_groups_api_report(
    client: NucleiClient, payload: dict, verbose: bool = False
) -> bytes:
    """
    Wrapper around the PileCore endpoint "/grouper/generate_grouper_report".

    Parameters
    ----------
    client: NucleiClient
        client object created by [nuclei](https://github.com/cemsbv/nuclei)
    payload: dict
        the payload of the request, can be created by calling `create_grouper_report_payload()`
    verbose: bool
        if True, print additional information to the console
    """
    logging.info(
        "Generate report... \n"
        "Depending on the amount of pile tip levels and CPT's this can take a while."
    )
    ticket = client.call_endpoint(
        "PileCore",
        "/grouper/generate_grouper_report",
        version="v3",
        schema=payload,
        return_response=True,
    )

    wait_until_ticket_is_ready(client=client, ticket=ticket, verbose=verbose)

    return client.call_endpoint(
        "PileCore", "/get-task-result", version="v3", schema=ticket.json()
    )


def get_multi_cpt_api_result_tension(
    client: NucleiClient,
    payload: dict,
    verbose: bool = False,
    standard: Literal["NEN9997-1", "CUR236"] = "NEN9997-1",
) -> dict:
    """
    Wrapper around the PileCore endpoint "/tension/[nen or cur]/multiple-cpts/results".

    Parameters
    ----------
    client: NucleiClient
        client object created by [nuclei](https://github.com/cemsbv/nuclei)
    payload: dict
        the payload of the request, can be created by calling `create_grouper_payload()`
    verbose: bool
        if True, print additional information to the console
    standard: str
        Norm used to calculate bearing capacities
    """
    logging.info(
        "Calculating bearing capacities... \n"
        "Depending on the amount of pile tip levels and CPT's this can take a while."
    )
    if standard == "NEN9997-1":
        endpoint = "/tension/nen/multiple-cpt/results"
    else:
        endpoint = "/tension/cur/multiple-cpt/results"
        payload.pop("construction_sequence", None)

    ticket = client.call_endpoint(
        "PileCore",
        endpoint=endpoint,
        version="v3",
        schema=payload,
        return_response=True,
    )

    wait_until_ticket_is_ready(client=client, ticket=ticket, verbose=verbose)

    return client.call_endpoint(
        "PileCore", "/get-task-result", version="v3", schema=ticket.json()
    )


def get_multi_cpt_api_report_tension(
    client: NucleiClient,
    payload: dict,
    verbose: bool = False,
    standard: Literal["NEN9997-1", "CUR236"] = "NEN9997-1",
) -> bytes:
    """
    Wrapper around the PileCore endpoint "/tension/[nen or cur]/multiple-cpts/report".

    Parameters
    ----------
    client: NucleiClient
        client object created by [nuclei](https://github.com/cemsbv/nuclei)
    payload: dict
        the payload of the request, can be created by calling `create_grouper_payload()`
    verbose: bool
        if True, print additional information to the console
    standard: str
        Norm used to calculate bearing capacities
    """
    logging.info(
        "Generate report... \n"
        "Depending on the amount of pile tip levels and CPT's this can take a while."
    )

    if standard == "NEN9997-1":
        endpoint = "/tension/nen/multiple-cpt/report"
    else:
        endpoint = "/tension/cur/multiple-cpt/report"
        payload.pop("construction_sequence", None)

    ticket = client.call_endpoint(
        "PileCore",
        endpoint=endpoint,
        version="v3",
        schema=payload,
        return_response=True,
    )
    wait_until_ticket_is_ready(client=client, ticket=ticket, verbose=verbose)

    return client.call_endpoint(
        "PileCore", "/get-task-result", version="v3", schema=ticket.json()
    )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, Literal, Tuple

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from numpy.typing import NDArray


def get_component_bounds_nap(
    pile_tip_level_nap: float | int,
    pile_head_level_nap: float | int,
    component_primary_length: float | int | None,
) -> Tuple[float, float]:
    """
    Returns component head and tip level in NAP.

    Parameters
    ----------
    pile_tip_level_nap : float
        pile tip level in [m] w.r.t. NAP.
    pile_head_level_nap : float
        pile head level in [m] w.r.t. NAP.
    component_primary_length : float, optional
        The length of the pile-geometry component.

    Returns
    -------
    tuple
        Tuple with component head and tip level in [m] w.r.t. NAP.
    """
    # Component tip level is always the pile tip level
    component_tip_level_nap = pile_tip_level_nap

    # Component head level is the pile head level if the length is not set
    if component_primary_length is not None:
        component_head_level_nap = pile_tip_level_nap + component_primary_length
    else:
        component_head_level_nap = pile_head_level_nap

    return float(component_head_level_nap), float(component_tip_level_nap)


def get_circum_vs_depth(
    depth_nap: NDArray[np.floating],
    pile_tip_level_nap: float | int,
    pile_head_level_nap: float | int,
    length: float | None,
    circumference: float,
) -> NDArray[np.floating]:
    """
    Returns component circumferences at requested depths.

    Parameters
    ----------
    depth_nap : np.array
        Array with depths in [m] w.r.t. NAP.
    pile_tip_level_nap : float
        pile tip level in [m] w.r.t. NAP.
    pile_head_level_nap : float
        pile head level in [m] w.r.t. NAP.
    length : float, optional
        The length of the pile-geometry component, by default None.
    circumference : float
        The circumference of the pile-geometry component.

    Returns
    -------
    np.array
        Array with component circumferences at the depths in the depth parameter.
    """
    circum_vs_depth = np.zeros_like(depth_nap)

    # Component tip level is always the pile tip level
    component_tip_level_nap = pile_tip_level_nap

    # Component head level is the pile head level if the length is not set
    if length is not None:
        component_head_level_nap = pile_tip_level_nap + length
    else:
        component_head_level_nap = pile_head_level_nap

    # Fill the circumference between component tip and head level
    circum_vs_depth[
        (depth_nap <= component_head_level_nap) & (depth_nap >= component_tip_level_nap)
    ] = circumference

    return circum_vs_depth


def get_area_vs_depth(
    depth_nap: NDArray[np.floating],
    area_full: float | int,
    component_head_level_nap: float | int,
    component_tip_level_nap: float | int,
    inner_area: NDArray[np.floating],
) -> NDArray[np.floating]:
    """
    Returns component areas at requested depths.

    Parameters
    ----------
    depth_nap : np.array
        Array with depths in [m] w.r.t. NAP.
    area_full : float
        The full outer-area [m²] of the pile-geometry component, including any potential inner-components.
    pile_tip_level_nap : float
        pile tip level in [m] w.r.t. NAP.
    pile_head_level_nap : float
        pile head level in [m] w.r.t. NAP.
    component_head_level_nap : float
        Component head level in [m] w.r.t. NAP.
    component_tip_level_nap : float
        Component tip level in [m] w.r.t. NAP.
    inner_area : np.array
        Array with inner areas at the depths in the depth parameter.

    Returns
    -------
    np.array
        Array with component areas at the depths in the depth parameter.
    """
    area_vs_depth = np.zeros_like(depth_nap, dtype=np.floating)

    # Mask the depths between the component tip and head level
    mask_depths = (depth_nap <= component_head_level_nap) & (
        depth_nap >= component_tip_level_nap
    )

    # Fill the area between component tip and head level, subtracting the inner area
    area_vs_depth[mask_depths] = float(area_full) - inner_area[mask_depths]

    return area_vs_depth


def instantiate_axes(
    figsize: Tuple[float, float] = (6.0, 6.0),
    axes: Axes | None = None,
    **kwargs: Any,
) -> Axes:
    """
    Validate axes objects if provided, otherwise create a new axes object.

    Parameters
    ----------
    figsize : tuple, optional
        The figure size (width, height) in inches, by default (6.0, 6.0).
    axes : Axes, optional
        The axes object to plot the cross-section on.
    **kwargs
        Additional keyword arguments to pass to the `plt.subplots()` function.

    Returns
    -------
    Axes
        The axes object to plot the cross-section on.
    """
    # Create axes objects if not provided
    if axes is not None:
        if not isinstance(axes, Axes):
            raise ValueError(
                "'axes' argument must be a `pyplot.axes.Axes` object or None."
            )

    else:
        kwargs_subplot = {
            "figsize": figsize,
            "tight_layout": True,
        }

        kwargs_subplot.update(kwargs)

        _, axes = plt.subplots(
            1,
            1,
            **kwargs_subplot,
        )

        if not isinstance(axes, Axes):
            raise ValueError(
                "Could not create Axes objects. This is probably due to invalid matplotlib keyword arguments. "
            )

    return axes


class _BasePileGeometryComponent(ABC):
    """
    The _BasePileGeometryComponent class is an abstract base class for pile-geometry components.
    """

    @classmethod
    @abstractmethod
    def from_api_response(
        cls,
        component: dict,
        inner_component: _BasePileGeometryComponent | None = None,
    ) -> _BasePileGeometryComponent:
        """
        Instantiates a pile-geometry component from a component object in the API response payload.

        Parameters:
        -----------
        component: dict
            A dictionary that is retrieved from the API response payload at "/pile_properties/geometry/components/[i]".
        inner_component: _BasePileGeometryComponent | None, optional
            The component on the inside of the pile-geometry component, by default None.

        Returns:
        --------
        _BasePileGeometryComponent
            A pile-geometry component.
        """
        ...

    @property
    @abstractmethod
    def inner_component(
        self,
    ) -> _BasePileGeometryComponent | None:
        """The component on the inside of the pile-geometry component"""
        ...

    @property
    @abstractmethod
    def outer_shape(self) -> str:
        """The outer shape of the pile-geometry component"""
        ...

    @property
    @abstractmethod
    def material(self) -> str | None:
        """The material name of the pile-geometry component"""
        ...

    @property
    @abstractmethod
    def primary_dimension(self) -> PrimaryPileComponentDimension:
        """
        The primary dimension [m] of the pile-geometry component, which is measured along the primary axis of the pile.
        """
        ...

    @property
    @abstractmethod
    def circumference(self) -> float:
        """The outer-circumference [m] of the pile-geometry component"""
        ...

    @property
    @abstractmethod
    def equiv_tip_diameter(self) -> float:
        """
        Equivalent outer-diameter [m] of the component at the tip-level.
        """
        ...

    @property
    @abstractmethod
    def area_full(self) -> float:
        """The full outer-area [m²] of the pile-geometry component, including any potential inner-components"""
        ...

    @abstractmethod
    def serialize_payload(self) -> dict:
        """
        Serialize the pile-geometry component to a dictionary payload for the API.

        Returns:
            A dictionary payload containing the geometry properties.
        """

    @abstractmethod
    def get_component_bounds_nap(
        self,
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
    ) -> Tuple[float, float]:
        """
        Returns component head and tip level in NAP.

        Parameters
        ----------
        pile_tip_level_nap : float
            pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            pile head level in [m] w.r.t. NAP.

        Returns
        -------
        tuple
            Tuple with component head and tip level in [m] w.r.t. NAP.
        """
        ...

    @abstractmethod
    def get_circum_vs_depth(
        self,
        depth_nap: NDArray[np.floating],
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
    ) -> NDArray[np.floating]:
        """
        Returns component circumferences at requested depths.

        Parameters
        ----------
        depth_nap : np.array
            Array with depths in [m] w.r.t. NAP.
        pile_tip_level_nap : float
            pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            pile head level in [m] w.r.t. NAP.

        Returns
        -------
        np.array
            Array with component circumferences at the depths in the depth parameter.
        """
        ...

    @abstractmethod
    def get_inner_area_vs_depth(
        self,
        depth_nap: NDArray[np.floating],
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
    ) -> NDArray[np.floating]:
        """
        Returns inner component areas at requested depths.

        Parameters
        ----------
        depth_nap : np.array
            Array with depths in [m] w.r.t. NAP.
        pile_tip_level_nap : float
            pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            pile head level in [m] w.r.t. NAP.

        Returns
        -------
        np.array
            Array with inner component areas at the depths in the depth parameter.
        """
        ...

    @abstractmethod
    def get_area_vs_depth(
        self,
        depth_nap: NDArray[np.floating],
        pile_tip_level_nap: float | int,
        pile_head_level_nap: float | int,
    ) -> NDArray[np.floating]:
        """
        Returns component areas at requested depths.

        Parameters
        ----------
        depth_nap : np.array
            Array with depths in [m] w.r.t. NAP.
        pile_tip_level_nap : float
            pile tip level in [m] w.r.t. NAP.
        pile_head_level_nap : float
            pile head level in [m] w.r.t. NAP.

        Returns
        -------
        np.array
            Array with component areas at the depths in the depth parameter.
        """
        ...

    @abstractmethod
    def plot_cross_section_exterior(
        self,
        figsize: Tuple[float, float] = (6.0, 6.0),
        facecolor: Tuple[float, float, float] | str | None = None,
        axes: Axes | None = None,
        axis_arg: bool | str | Tuple[float, float, float, float] | None = "auto",
        show: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the cross-section of the component at a specified depth.

        Parameters
        ----------
        figsize : tuple, optional
            The figure size (width, height) in inches, by default (6.0, 6.0).
        facecolor : tuple or str, optional
            The face color of the cross-section, by default None.
        axes : Axes
            The axes object to plot the cross-section on.
        axis_arg : bool or str or tuple, optional
            The axis argument to pass to the `axes.axis()` function, by default "auto".
        show : bool, optional
            Whether to display the plot, by default True.

        Returns
        -------
        Axes
            The axes object to plot the cross-section on.
        """
        ...

    @abstractmethod
    def plot_side_view(
        self,
        bottom_boundary_nap: float | Literal["pile_tip"] = "pile_tip",
        top_boundary_nap: float | Literal["pile_head"] = "pile_head",
        pile_tip_level_nap: float | int = -10,
        pile_head_level_nap: float | int = 0,
        figsize: Tuple[float, float] = (6.0, 6.0),
        facecolor: Tuple[float, float, float] | str | None = None,
        axes: Axes | None = None,
        axis_arg: bool | str | Tuple[float, float, float, float] | None = "scaled",
        show: bool = True,
        **kwargs: Any,
    ) -> Axes:
        """
        Plot the side view of the cross-section at a specified depth.

        Parameters
        ----------
        figsize : tuple, optional
            The figure size (width, height) in inches, by default (6.0, 6.0).
        facecolor : tuple or str, optional
            The face color of the pile cross-section, by default None.
        axes : Axes
            The axes object to plot the cross-section on.
        axis_arg : bool or str or tuple, optional
            The axis argument to pass to the `axes.axis()` function, by default "auto".
        show : bool, optional
            Whether to display the plot, by default True.
        **kwargs
            Additional keyword arguments to pass to the `plt
        """
        ...


class PrimaryPileComponentDimension:
    """
    The PrimaryPileComponentDimension class represents the primary dimension of a pile-geometry component,
    which is measured along the primary axis of the pile. It contains the length, top level, and bottom level
    of the pile-geometry component.
    """

    def __init__(
        self,
        length: float | None = None,
    ):
        """
        Represents the primary dimension of a pile-geometry component, which is measured
        along the primary axis of the pile.

        Args:
            length: The length [m] of the pile-geometry component (along the primary axis).
        """
        self._length = length

    @classmethod
    def from_api_response(cls, primary_dim: dict) -> PrimaryPileComponentDimension:
        """
        Instantiates a PrimaryPileComponentDimension from a primary dimension object in the API response payload.

        Args:
            primary_dim: A dictionary that is retrieved from the API response payload at "/pile_properties/geometry/components/[i]/primary_dimension".
        """
        return cls(
            length=primary_dim.get("length"),
        )

    @property
    def length(self) -> float | None:
        """The length [m] of the pile-geometry component"""
        return self._length

    def serialize_payload(self) -> Dict[str, float | None] | None:
        """
        Serialize the primary dimension to a dictionary payload for the API.

        Only contains the length, since this is the only accepted value by the API.
        Returns None if the length is not set.

        Returns:
            A dictionary payload containing the length of the primary dimension.
        """
        if self.length is not None:
            return {"length": self.length}
        return None



from typing import Any

import numpy as np
import plotly


def get_continuous_color(colorscale: Any, intermed: float) -> Any:
    """
    Plotly continuous colorscales assign colors to the range [0, 1]. This function computes the intermediate
    color for any value in that range.

    Plotly doesn't make the colorscales directly accessible in a common format.
    Some are ready to use:

        colorscale = plotly.colors.PLOTLY_SCALES["Greens"]

    Others are just swatches that need to be constructed into a colorscale:

        viridis_colors, scale = plotly.colors.convert_colors_to_same_type(plotly.colors.sequential.Viridis)
        colorscale = plotly.colors.make_colorscale(viridis_colors, scale=scale)

    :param colorscale: A plotly continuous colorscale defined with RGB string colors.
    :param intermed: value in the range [0, 1]. If `intermed` is NaN, it returns the first color in the colorscale.
    :return: color in rgb string format
    :rtype: str
    """
    if len(colorscale) < 1:
        raise ValueError("colorscale must have at least one color")

    if intermed <= 0 or len(colorscale) == 1 or np.isnan(intermed):
        return colorscale[0][1]
    if intermed >= 1:
        return colorscale[-1][1]

    for cutoff, color in colorscale:
        if intermed > cutoff:
            low_cutoff, low_color = cutoff, color
        else:
            high_cutoff, high_color = cutoff, color
            break

    # noinspection PyUnboundLocalVariable
    return plotly.colors.find_intermediate_color(
        lowcolor=low_color,
        highcolor=high_color,
        intermed=((intermed - low_cutoff) / (high_cutoff - low_cutoff)),
        colortype="rgb",
    )
